#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: JM
"""

import gurobipy as gp
from forest import *
from tree import *
from domain import *
from instance import *
import time



class Contrastive:
    """
    Class used to compute a contrastive explaination.
    """

    def __init__(self, timeLimit = 100.0) -> None:
        """
        Constructor.

        """
        self.model = gp.Model()
        self.model.setParam(gp.GRB.Param.OutputFlag, 0)
        self.model.setParam(gp.GRB.Param.Threads, 1)
        self.model.setParam(gp.GRB.Param.Method, 0)
        self.model.setParam(gp.GRB.Param.MIPGap, 0.01)
        self.model.setParam(gp.GRB.Param.TimeLimit, timeLimit)
        self.model.setParam(gp.GRB.Param.Threads, 1)
        self.model.setParam(gp.GRB.Param.MIPFocus, 1)        
        self.model.setParam(gp.GRB.Param.MIPGapAbs, 1e-2)
        self.model.setParam(gp.GRB.Param.MIPGap, 1e-2)
        self.model.setParam(gp.GRB.Param.Heuristics, 0.0)
        self.model.setParam(gp.GRB.Param.Presolve, 2.0)     
        self.model.setParam(gp.GRB.Param.FeasibilityTol, 1e-9)
        
        self.varTrees = []
        self.varActiveBranches = []
        self.varValueTree = []
        self.varValueForest = None

    def constructForestModel(self, forest: Forest) -> None:
        """
        Construct the basic model used to represent the dynamic of the boosted tree machine learning model.
        Args:
            forest (Forest): the forest we want to encode.
        """
        # Create variables
        self.varTrees = [self.model.addVar(vtype=gp.GRB.BINARY, name="b" + str(i))
                         for i in range(1 + forest.getMaxIndex())]
        self.varActiveBranches = [self.model.addVar(vtype=gp.GRB.BINARY, name='a' + str(i))
                                  for i in range(forest.getNbAllBranches())]
        self.varValueTree = [self.model.addVar(vtype=gp.GRB.CONTINUOUS, lb=t.getMinValue(
        ), ub=t.getMaxValue(), name="t" + str(t.getId())) for t in forest]
        self.varValueForest = self.model.addVar(vtype=gp.GRB.CONTINUOUS, lb=forest.getMinPossiblePrediction(
        ), ub=forest.getMaxPossiblePrediction(), name="forest")

        # Add constraints about the trees.
        cpt = 0
        for tree in forest:
            self.model.addLConstr(gp.LinExpr(
                [1 for _ in tree], [self.varActiveBranches[branch.getId()] for branch in tree]), gp.GRB.EQUAL, 1)

            self.model.addLConstr(gp.LinExpr([branch.getValue() for branch in tree], [
                                  self.varActiveBranches[branch.getId()] for branch in tree]), gp.GRB.EQUAL, self.varValueTree[tree.getId()])

            for branch in tree:
                coeff = [-1]
                lits = [self.varActiveBranches[branch.getId()]]
                nbNeg = 0
                for l in branch.getLiterals():
                    if l > 0:
                        lits.append(self.varTrees[l])
                        coeff.append(1)
                    else:
                        lits.append(self.varTrees[-l])
                        coeff.append(-1)
                        nbNeg += 1

                self.model.addLConstr(gp.LinExpr(coeff, lits), gp.GRB.LESS_EQUAL, len(
                    branch.getLiterals()) - 1 - nbNeg)
            
            
        # Add a constraint to get the value for the forest.
        c = self.model.addLConstr(gp.LinExpr([1 for _ in self.varValueTree], [
                              v for v in self.varValueTree]), gp.GRB.EQUAL, self.varValueForest)
        

    def constructTheoryModel(self, clauses):
        """
        Add a set of constraint to the model in order to take into account the domain constraints.
        Args:
            forest: Forest
                is the BT we want to explain the decision.
            theory: list(list(int))
                is a set of clauses imposing constraints on possible combinaison of binary features.
        """
        for cl in clauses:
            coeff = []
            lits = []
            nbNeg = 0
            for l in cl:
                if l > 0:
                    coeff.append(1)
                    lits.append(self.varTrees[l])
                else:
                    coeff.append(-1)
                    lits.append(self.varTrees[-l])
                    nbNeg += 1
            self.model.addLConstr(gp.LinExpr(coeff, lits),
                                  gp.GRB.GREATER_EQUAL, 1 - nbNeg)
            
    def constructOptimFunction(self, domains : Domain, instance : Instance):
        """
        Construct the constraint needed for handling the user preferences. This function returns the linear function we want to minimize.

        Args:
            domains : Domain
                gives information about the feature (type and associated boolean variables).
            instance : Instance
                gives the instance we want to explain.
        """
        coeffs = []
        lits = []
        for feature in instance.scoringMap:
            assert feature in domains.mapping
            var = domains.mapping[feature]
            if var.getType() == VariableType.CATEGORICAL:                
                for value in var.values:  
                    if instance.literals[value - 1] == 1: # we do not consider when that does not change.
                        continue
                        
                    w = instance.getWeightCategorical(feature, value)
                    
                    if type(w) == str:
                        assert w == 'inf'
                        self.model.addConstr(self.varTrees[value] == 0)
                    else:
                        coeffs.append(w)
                        lits.append(self.varTrees[value])                    
            else:                
                var.threshold = sorted(var.threshold, key = lambda x : x[1])
                if instance.literals[var.threshold[0][0] - 1] == 1:                    
                    w = instance.getWeightNumerical(feature, '-inf', var.threshold[0][1])
                    if type(w) == str:
                        assert w == 'inf'
                        self.model.addConstr(self.varTrees[var.threshold[0][0]] == 1)
                    else:
                        coeffs.append(-w)
                        lits.append(self.varTrees[var.threshold[0][0]])
                
                if instance.literals[var.threshold[-1][0] - 1] == 0:
                    w = instance.getWeightNumerical(feature, var.threshold[-1][1], 'inf')
                    if type(w) == str:
                        assert w == 'inf'
                        self.model.addConstr(self.varTrees[var.threshold[-1][0]] == 0)
                    else:                        
                        coeffs.append(w)
                        lits.append(self.varTrees[var.threshold[-1][0]])
                
                for i in range(len(var.threshold) - 1):
                    if instance.literals[var.threshold[i][0] - 1] == 1 and instance.literals[var.threshold[i+1][0] - 1] == 0:
                        continue
                    
                    w = instance.getWeightNumerical(feature, var.threshold[i][1], var.threshold[i+1][1])
                    vw = self.model.addVar(vtype=gp.GRB.BINARY, name=f"w_{var.name}_[{var.threshold[i][1]}, {var.threshold[i+1][1]}[")
                    vwr = self.model.addVar(vtype=gp.GRB.BINARY, name=f"w_{var.name}_r_[{var.threshold[i][1]}, {var.threshold[i+1][1]}[")
                    
                    # link vw with the interval.
                    self.model.addConstr(self.varTrees[var.threshold[i][0]] + self.varTrees[var.threshold[i+1][0]] == 2 * vwr + vw)
                    
                    if type(w) == str:
                        assert w == 'inf'
                        self.model.addConstr(vw == 0)
                    else:
                        coeffs.append(w)
                        lits.append(vw)
                    
        self.model.update() 
        return coeffs, lits
    

    def getSolution(self):
        """
        Display the solution.
        """
        if self.model.SolCount == 0:
            return []

        interpretation = [0]
        for i in range(1, len(self.varTrees)):
            if self.varTrees[i].X > 0.9:
                interpretation.append(i)
            else:
                interpretation.append(-i)
                        
        return interpretation
    
    
    def runMinMaxInterval(self, forest: Forest, domains: Domain):
        """
        Compute the min/max possible value for the forest.
        Args:
            forest (Forest): the forest we want to explain the decision.
            theory (list(list(int))): the set of clauses encoding the theory.            
        """
        start = time.time()        
        
        # construct the model for the forest and domain theory.
        self.constructForestModel(forest)
        self.constructTheoryModel(domains.generateTheoryClauses())

        self.model.setObjective(self.varValueForest, gp.GRB.MINIMIZE)
        self.model.optimize()
        
        if self.model.Status == gp.GRB.TIME_LIMIT:
            return ('to', 'na', 'na', 'na')
        
        minBound = self.varValueForest.X
        print("c min:", minBound)
        
        self.model.setObjective(self.varValueForest, gp.GRB.MAXIMIZE)
        self.model.optimize()
        
        if self.model.Status == gp.GRB.TIME_LIMIT:
            return ('to', 'na', 'na', 'na')        
        
        maxBound = self.varValueForest.X
        print("c max:", maxBound)
                
        end = time.time()
        
        return ('solved', minBound, maxBound, end - start)


            
    def run(self, forest: Forest, domains: Domain, instance : Instance, lb : float, ub : float) -> list:
        """
        Compute a contrastive explanation.
        Args:
            forest (Forest): the forest we want to explain the decision.
            theory (list(list(int))): the set of clauses encoding the theory.
            instances: the instance we want to explain.
        """
        start = time.time()        
        
        # construct the model for the forest and domain theory.
        self.constructForestModel(forest)
        self.constructTheoryModel(domains.generateTheoryClauses())

        
        # set the interval.
        self.varValueForest.setAttr('lb', lb)
        self.varValueForest.setAttr('ub', ub)

        # construct the function we want to minimize.
        coeffs, lits = self.constructOptimFunction(domains, instance)
        self.model.setObjective(gp.LinExpr(coeffs, lits), gp.GRB.MINIMIZE)
        
        # to print the model
        # self.model.write("out.lp")        
        
        # compute a solution
        self.model.optimize()
        end = time.time()
        
        if self.model.Status == gp.GRB.TIME_LIMIT:
            return ('to', end - start, 'na', [])
        
        interpretation = self.getSolution()
                                
        status = 'unsat'
        contrastivePrediction = 'na'
        
        if len(interpretation):
            status = 'sat'
            contrastivePrediction = forest.computePrediction(interpretation)        
            if contrastivePrediction < lb or contrastivePrediction > ub:
                print(f"c error {contrastivePrediction} is not in [{lb},{ub}]")            
                    
        return (status, end - start, contrastivePrediction, interpretation[1:])

